@@ -1,6 +1,6 @@
 /*
  * Pound - the reverse-proxy load-balancer
- * Copyright (C) 2023 Sergey Poznyakoff
+ * Copyright (C) 2023-2025 Sergey Poznyakoff
  *
  * Pound is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -262,11 +262,39 @@ save_forwarded_header (POUND_HTTP *phttp)
   char const *val;
 
   free (phttp->orig_forwarded_header);
+  phttp->orig_forwarded_header = NULL;
   hname = get_forwarded_header_name (phttp);
   if ((hdr = http_header_list_locate_name (&phttp->request.headers,
-					   hname, strlen (hname))) != NULL
-      && (val = http_header_get_value (hdr)) != NULL)
-    phttp->orig_forwarded_header = xstrdup (val);
+					   hname, strlen (hname))) != NULL)
+    {
+      if (is_combinable_header (hdr))
+	{
+	  if ((val = http_header_get_value (hdr)) != NULL)
+	    phttp->orig_forwarded_header = xstrdup (val);
+	}
+      else
+	{
+	  struct stringbuf sb;
+
+	  stringbuf_init_log (&sb);
+	  for (;;)
+	    {
+	      size_t len = hdr->val_end - hdr->val_start;
+
+	      stringbuf_add (&sb, hdr->header + hdr->val_start, len);
+	      if ((hdr = http_header_list_next (hdr)) != NULL)
+		{
+		  if (len > 0)
+		    stringbuf_add (&sb, ", ", 2);
+		}
+	      else
+		break;
+	    }
+	  phttp->orig_forwarded_header = stringbuf_finish (&sb);
+	  if (!phttp->orig_forwarded_header)
+	    stringbuf_free (&sb);
+	}
+    }
 }
 
 static void
@@ -622,7 +650,7 @@ be_service_name (BACKEND *be)
 {
   switch (be->be_type)
     {
-    case BE_BACKEND:
+    case BE_REGULAR:
       if (be->service->name)
        return be->service->name;
       break;
@@ -636,6 +664,7 @@ be_service_name (BACKEND *be)
       return "(error)";
     case BE_METRICS:
       return "(metrics)";
+    case BE_MATRIX:
     case BE_BACKEND_REF:
       /* shouldn't happen */
       break;
@@ -682,21 +711,21 @@ static void
 i_backend_locus (struct stringbuf *sb, struct http_log_instr *instr,
 		 POUND_HTTP *phttp)
 {
-  print_str (sb, phttp->backend->locus);
+  print_str (sb, phttp->backend->locus_str);
 }
 
 static void
 i_service_locus (struct stringbuf *sb, struct http_log_instr *instr,
 		 POUND_HTTP *phttp)
 {
-  print_str (sb, phttp->svc->locus);
+  print_str (sb, phttp->svc->locus_str);
 }
 
 static void
 i_listener_locus (struct stringbuf *sb, struct http_log_instr *instr,
 		 POUND_HTTP *phttp)
 {
-  print_str (sb, phttp->lstn->locus);
+  print_str (sb, phttp->lstn->locus_str);
 }
 
 static struct argprt locprt[] = {
