@@ -1,7 +1,7 @@
 /*
  * Pound - the reverse-proxy load-balancer
  * Copyright (C) 2002-2010 Apsis GmbH
- * Copyright (C) 2018-2023 Sergey Poznyakoff
+ * Copyright (C) 2018-2025 Sergey Poznyakoff
  *
  * Pound is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -43,13 +43,20 @@ LISTENER_HEAD listeners = SLIST_HEAD_INITIALIZER (listeners);
 				/* all available listeners */
 int n_listeners;                /* Number of listeners */
 
-regex_t HEADER,			/* Allowed header */
+GENPAT HEADER,		/* Allowed header */
   CONN_UPGRD,			/* upgrade in connection header */
   LOCATION;			/* the host we are redirected to */
 
 char *forwarded_header;         /* "forwarded" header name */
 ACL *trusted_ips;               /* Trusted IP addresses */
 
+char *syslog_tag;               /* Tag to mark syslog messages with.
+				   If NULL, progname will be used. */
+
+pthread_mutexattr_t mutex_attr_recursive;
+pthread_attr_t thread_attr_detached;
+static pthread_attr_t thread_attr_worker;
+
 #ifndef  SOL_TCP
 /* for systems without the definition */
 int SOL_TCP;
@@ -103,6 +110,24 @@ logmsg (const int priority, const char *fmt, ...)
   va_end (ap);
 }
 
+/*
+ * Log an out of memory condition and return.
+ * Take care not to use [v]logmsg as this could create infinite recursion.
+ */
+void
+lognomem (void)
+{
+  if (log_facility == -1 || print_log)
+    {
+      fprintf (stderr, "%s: out of memory\n", progname);
+    }
+
+  if (log_facility != -1)
+    {
+      syslog (LOG_CRIT, "out of memory");
+    }
+}
+
 /*
  * This is used as exit point if memory allocation failures occur at program
  * startup (e.g. when parsing config or the like).
@@ -110,20 +135,10 @@ logmsg (const int priority, const char *fmt, ...)
 void
 xnomem (void)
 {
-  logmsg (LOG_CRIT, "out of memory");
+  lognomem ();
   exit (1);
 }
 
-/*
- * This is used as json_memabrt hook.  The code in svc.c handles memory
- * allocation failures gracefully and returns 500 if any occurs.
- */
-void
-lognomem (void)
-{
-  logmsg (LOG_CRIT, "out of memory");
-}
-
 /*
  * Log a message at LOG_CRIT and terminate the program.
  */
@@ -198,7 +213,6 @@ static unsigned active_threads; /* Number of threads currently active,
 				   i.e processing requests. */
 
 static unsigned worker_count = 0;
-static pthread_attr_t attr;
 
 unsigned worker_min_count = DEFAULT_WORKER_MIN;
 unsigned worker_max_count = DEFAULT_WORKER_MAX;
@@ -238,7 +252,7 @@ worker_start (void)
   pthread_t thr;
   int rc;
 
-  if ((rc = pthread_create (&thr, &attr, thr_http, NULL)) != 0)
+  if ((rc = pthread_create (&thr, &thread_attr_worker, thr_http, NULL)) != 0)
     abend ("can't create worker thread: %s", strerror (rc));
   worker_count++;
 }
@@ -319,7 +333,7 @@ pound_http_dequeue (void)
 	    {
 	      /*
 	       * worker_count might have changed while we were waiting,
-	       * so check again if the mimnimal worker count is reached.
+	       * so check again if the minimal worker count is reached.
 	       */
 	      if (worker_count == worker_min_count)
 		continue;
@@ -375,6 +389,7 @@ pound_http_destroy (POUND_HTTP *arg)
       BIO_ssl_shutdown (arg->cl);
     }
 
+  backend_unref (arg->backend);
   if (arg->be != NULL)
     {
       BIO_flush (arg->be);
@@ -697,15 +712,9 @@ server (void)
   pthread_sigmask (SIG_BLOCK, &sigs, NULL);
 
   /* thread stuff */
-  pthread_attr_init (&attr);
-  pthread_attr_setdetachstate (&attr, PTHREAD_CREATE_DETACHED);
-
-  /* set new stack size  */
-  if (pthread_attr_setstacksize (&attr, 1 << 18))
-    abend ("can't set stack size");
 
   /* start timer */
-  if (pthread_create (&thr, &attr, thr_timer, NULL))
+  if (pthread_create (&thr, &thread_attr_detached, thr_timer, NULL))
     abend ("can't create timer thread: %s", strerror (errno));
 
   /*
@@ -954,6 +963,146 @@ probe_pthread_cancel (void)
 # define probe_pthread_cancel()
 #endif
 
+static int
+read_fd (int fd)
+{
+  struct msghdr msg;
+  struct iovec iov[1];
+  char base[1];
+  union
+  {
+    struct cmsghdr cm;
+    char control[CMSG_SPACE (sizeof (int))];
+  } control_un;
+  struct cmsghdr *cmptr;
+
+  msg.msg_control = control_un.control;
+  msg.msg_controllen = sizeof (control_un.control);
+
+  msg.msg_name = NULL;
+  msg.msg_namelen = 0;
+
+  iov[0].iov_base = base;
+  iov[0].iov_len = sizeof (base);
+
+  msg.msg_iov = iov;
+  msg.msg_iovlen = 1;
+  if (recvmsg (fd, &msg, 0) > 0)
+    {
+      if ((cmptr = CMSG_FIRSTHDR (&msg)) != NULL
+	  && cmptr->cmsg_len == CMSG_LEN (sizeof (int))
+	  && cmptr->cmsg_level == SOL_SOCKET
+	  && cmptr->cmsg_type == SCM_RIGHTS)
+	return *((int*) CMSG_DATA (cmptr));
+    }
+  return -1;
+}
+
+static void
+listener_socket_from (LISTENER *lst)
+{
+  struct sockaddr_storage ss;
+  socklen_t sslen = sizeof (ss);
+  int sfd, fd;
+  struct stringbuf sb;
+  char tmp[MAX_ADDR_BUFSIZE];
+
+  if ((sfd = socket (PF_UNIX, SOCK_STREAM, 0)) < 0)
+    abend ("%s: socket: %s", lst->locus_str, strerror (errno));
+
+  if (connect (sfd, lst->addr.ai_addr, lst->addr.ai_addrlen) < 0)
+    abend ("%s: connect %s: %s",
+	   lst->locus_str,
+	   ((struct sockaddr_un*)lst->addr.ai_addr)->sun_path,
+	   strerror (errno));
+
+  fd = read_fd (sfd);
+
+  if (fd == -1)
+    abend ("%s: can't get socket: %s", lst->locus_str, strerror (errno));
+
+  if (getsockname (fd, (struct sockaddr*) &ss, &sslen) == -1)
+    abend ("%s: can't get socket address: %s",
+	   lst->locus_str, strerror (errno));
+
+  lst->sock = fd;
+  
+  free (lst->addr.ai_addr);
+  lst->addr.ai_addr = xmalloc (sslen);
+  memcpy (lst->addr.ai_addr, &ss, sslen);
+  lst->addr.ai_addrlen = sslen;
+  lst->addr.ai_family = ss.ss_family;
+
+  xstringbuf_init (&sb);
+  stringbuf_add_string (&sb, lst->locus_str);
+  stringbuf_add_string (&sb, ": obtained address ");
+  stringbuf_add_string (&sb, addr2str (tmp, sizeof (tmp), &lst->addr, 0));
+  logmsg (LOG_DEBUG, "%s", stringbuf_finish (&sb));
+  stringbuf_free (&sb);
+}
+
+static void
+listener_init (LISTENER *lst, uid_t user_id, gid_t group_id)
+{
+  int opt;
+  int domain;
+  char abuf[MAX_ADDR_BUFSIZE];
+  mode_t oldmask;
+
+  switch (lst->addr.ai_family)
+    {
+    case AF_INET:
+      domain = PF_INET;
+      break;
+
+    case AF_INET6:
+      domain = PF_INET6;
+      break;
+
+    case AF_UNIX:
+      domain = PF_UNIX;
+      unlink (((struct sockaddr_un*)lst->addr.ai_addr)->sun_path);
+      oldmask = umask (0777 & ~lst->mode);
+      break;
+      
+    default:
+      abort ();
+    }
+
+  if ((lst->sock = socket (domain, SOCK_STREAM, 0)) < 0)
+    abend ("%s: can't create HTTP socket %s: %s",
+	   lst->locus_str,
+	   addr2str (abuf, sizeof (abuf), &lst->addr, 0),
+	   strerror (errno));
+
+  opt = 1;
+  setsockopt (lst->sock, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof (opt));
+  if (bind (lst->sock, lst->addr.ai_addr, lst->addr.ai_addrlen) < 0)
+    abend ("%s: can't bind HTTP socket to %s: %s",
+	   lst->locus_str,
+	   addr2str (abuf, sizeof (abuf), &lst->addr, 0),
+	   strerror (errno));
+  
+  if (domain == PF_UNIX)
+    {
+      umask (oldmask);
+      if (lst->chowner)
+	{
+	  char *sname = ((struct sockaddr_un*)lst->addr.ai_addr)->sun_path;
+	  if (chown (sname, user_id, group_id))
+	    logmsg (LOG_ERR, "%s: can't chown socket %s to %d:%d: %s",
+		    lst->locus_str,
+		    sname, user_id, group_id, strerror (errno));
+	}
+    }
+
+  if (listen (lst->sock, 512))
+    abend ("%s: can't listen on %s: %s",
+	   lst->locus_str,
+	   addr2str (abuf, sizeof (abuf), &lst->addr, 0),
+	   strerror (errno));
+}
+
 int
 main (const int argc, char **argv)
 {
@@ -976,12 +1125,15 @@ main (const int argc, char **argv)
   CRYPTO_set_locking_callback (l_lock);
 
   /* prepare regular expressions */
-  if (regcomp (&HEADER, "^([a-z0-9!#$%&'*+.^_`|~-]+):[ \t]*(.*)[ \t]*$",
-	       REG_ICASE | REG_NEWLINE | REG_EXTENDED)
-      || regcomp (&CONN_UPGRD, "(^|[ \t,])upgrade([ \t,]|$)",
-		  REG_ICASE | REG_NEWLINE | REG_EXTENDED)
-      || regcomp (&LOCATION, "(http|https)://([^/]+)(.*)",
-		  REG_ICASE | REG_NEWLINE | REG_EXTENDED))
+  if (genpat_compile (&HEADER, GENPAT_POSIX,
+		     "^([a-z0-9!#$%&'*+.^_`|~-]+):[ \t]*(.*)[ \t]*$",
+		     GENPAT_ICASE | GENPAT_MULTILINE)
+      || genpat_compile (&CONN_UPGRD, GENPAT_POSIX,
+			"(^|[ \t,])upgrade([ \t,]|$)",
+			GENPAT_ICASE | GENPAT_MULTILINE)
+      || genpat_compile (&LOCATION,  GENPAT_POSIX,
+			"(http|https)://([^/]+)(.*)",
+			GENPAT_ICASE | GENPAT_MULTILINE))
     abend ("bad essential Regex");
 
 #ifndef SOL_TCP
@@ -994,13 +1146,27 @@ main (const int argc, char **argv)
   SOL_TCP = pe->p_proto;
 #endif
 
+  pthread_mutexattr_init (&mutex_attr_recursive);
+  pthread_mutexattr_settype (&mutex_attr_recursive, PTHREAD_MUTEX_RECURSIVE);
+
+  pthread_attr_init (&thread_attr_detached);
+  pthread_attr_setdetachstate (&thread_attr_detached, PTHREAD_CREATE_DETACHED);
+
+  pthread_attr_init (&thread_attr_worker);
+  pthread_attr_setdetachstate (&thread_attr_worker, PTHREAD_CREATE_DETACHED);
+  /* set new stack size  */
+  if (pthread_attr_setstacksize (&thread_attr_worker, 1 << 18))
+    abend ("can't set stack size");
+
+  
   json_memabrt = lognomem;
 
   /* read config */
   config_parse (argc, argv);
 
   if (log_facility != -1)
-    openlog (progname, LOG_CONS | LOG_NDELAY | LOG_PID, log_facility);
+    openlog (syslog_tag ? syslog_tag : progname,
+	     LOG_CONS | LOG_NDELAY | LOG_PID, log_facility);
 
   /* set uid if necessary */
   if (user)
@@ -1028,60 +1194,10 @@ main (const int argc, char **argv)
   n_listeners = 0;
   SLIST_FOREACH (lstn, &listeners, next)
     {
-      if (lstn->sock == -1)
-	{
-	  /* prepare the socket */
-	  int opt;
-	  int domain;
-	  char abuf[MAX_ADDR_BUFSIZE];
-	  mode_t oldmask;
-
-	  switch (lstn->addr.ai_family)
-	    {
-	    case AF_INET:
-	      domain = PF_INET;
-	      break;
-	    case AF_INET6:
-	      domain = PF_INET6;
-	      break;
-	    case AF_UNIX:
-	      domain = PF_UNIX;
-	      unlink (((struct sockaddr_un*)lstn->addr.ai_addr)->sun_path);
-	      oldmask = umask (0777 & ~lstn->mode);
-	      break;
-	    default:
-	      abort ();
-	    }
-	  if ((lstn->sock = socket (domain, SOCK_STREAM, 0)) < 0)
-	    abend ("can't create HTTP socket %s: %s",
-		   addr2str (abuf, sizeof (abuf), &lstn->addr, 0),
-		   strerror (errno));
-
-	  opt = 1;
-	  setsockopt (lstn->sock, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof (opt));
-	  if (bind (lstn->sock, lstn->addr.ai_addr,
-		    (socklen_t) lstn->addr.ai_addrlen) < 0)
-	    abend ("can't bind HTTP socket to %s: %s",
-		   addr2str (abuf, sizeof (abuf), &lstn->addr, 0),
-		   strerror (errno));
-
-	  if (domain == PF_UNIX)
-	    {
-	      umask (oldmask);
-	      if (lstn->chowner)
-		{
-		  char *sname = ((struct sockaddr_un*)lstn->addr.ai_addr)->sun_path;
-		  if (chown (sname, user_id, group_id))
-		    logmsg (LOG_ERR, "can't chown socket %s to %d:%d: %s",
-			    sname, user_id, group_id, strerror (errno));
-		}
-	    }
-
-	  if (listen (lstn->sock, 512))
-	    abend ("can't listen on %s: %s",
-                   addr2str (abuf, sizeof (abuf), &lstn->addr, 0),
-                   strerror (errno));
-	}
+      if (lstn->socket_from)
+	listener_socket_from (lstn);
+      else
+	listener_init (lstn, user_id, group_id);
       n_listeners++;
     }
 
