@@ -1,7 +1,7 @@
 /*
  * Pound - the reverse-proxy load-balancer
  * Copyright (C) 2002-2010 Apsis GmbH
- * Copyright (C) 2018-2023 Sergey Poznyakoff
+ * Copyright (C) 2018-2025 Sergey Poznyakoff
  *
  * This file is part of Pound.
  *
@@ -69,6 +69,14 @@
 
 #define MAX_ADDR_BUFSIZE (NI_MAXHOST + NI_MAXSERV + 4)
 
+#define PORT_HTTP 80
+#define PORT_HTTPS 443
+
+#define TS(s) #s
+#define TOSTR(s) TS(s)
+#define PORT_HTTP_STR TOSTR(PORT_HTTP)
+#define PORT_HTTPS_STR TOSTR(PORT_HTTPS)
+
 #if HAVE_OPENSSL_SSL_H
 # define OPENSSL_THREAD_DEFINES
 # include <openssl/ssl.h>
@@ -91,32 +99,78 @@
 # include <openssl/engine.h>
 #endif
 
-#if HAVE_LIBPCREPOSIX == 2
-# include <pcre2posix.h>
-#elif HAVE_LIBPCREPOSIX == 1
-# if HAVE_PCREPOSIX_H
-#  include <pcreposix.h>
-# elif HAVE_PCRE_PCREPOSIX_H
-#  include <pcre/pcreposix.h>
-# else
-#  error "You have libpcreposix, but the header files are missing. Use --disable-pcreposix"
-# endif
+struct genpat
+{
+  struct genpat_defn const *vtab;
+  void *data;
+};
+
+typedef struct genpat *GENPAT;
+
+typedef struct {
+  int rm_so;
+  int rm_eo;
+} POUND_REGMATCH;
+
+#define GENPAT_DEFAULT   0
+#define GENPAT_ICASE     0x1
+#define GENPAT_MULTILINE 0x2
+
+int genpat_compile (GENPAT *, int, const char *, int);
+int genpat_match (GENPAT, const char *, size_t, POUND_REGMATCH *);
+void genpat_free (GENPAT);
+char const *genpat_error (GENPAT, size_t *);
+size_t genpat_nsub (GENPAT);
+
+enum
+  {
+    GENPAT_POSIX,
+    GENPAT_PCRE,
+    GENPAT_PREFIX,
+    GENPAT_SUFFIX,
+    GENPAT_CONTAIN,
+    GENPAT_EXACT,
+  };
+
+struct genpat_defn
+{
+  int (*gp_init) (GENPAT);
+  int (*gp_compile) (void *, const char *, int);
+  char const *(*gp_error) (void *, size_t *);
+  int (*gp_exec) (void *, const char *, size_t, POUND_REGMATCH *);
+  size_t (*gp_nsub) (void *);
+  void (*gp_free) (void *);
+};
+
+extern struct genpat_defn posix_genpat_defn;
+extern struct genpat_defn prefix_genpat_defn;
+extern struct genpat_defn suffix_genpat_defn;
+extern struct genpat_defn contain_genpat_defn;
+extern struct genpat_defn exact_genpat_defn;
+
+#ifdef HAVE_LIBPCRE
+extern struct genpat_defn pcre_genpat_defn;
+# define PCRE_REGEX_DEFN &pcre_genpat_defn
 #else
-# include <regex.h>
+# define PCRE_REGEX_DEFN NULL
 #endif
 
 #ifdef  HAVE_LONG_LONG_INT
 typedef long long CONTENT_LENGTH;
-# define STRTOCLEN strtoll
+# define CONTENT_LENGTH_MAX LLONG_MAX
 # define PRICLEN "lld"
 #else
 typedef long CONTENT_LENGTH;
-# define STRTOCLEN strtol
+# define CONTENT_LENGTH_MAX LONG_MAX
 # define PRICLEN "ld"
 #endif
 
 #define NO_CONTENT_LENGTH ((CONTENT_LENGTH) -1)
 
+int strtoclen (char const *arg, int base, CONTENT_LENGTH *retval, char **endptr);
+unsigned long strhash_ci (const char *c, size_t len);
+
+
 #ifndef DEFAULT_WORKER_MIN
 # define DEFAULT_WORKER_MIN 5
 #endif
@@ -210,6 +264,7 @@ enum
     HTTP_STATUS_UNAUTHORIZED,      // 401
     HTTP_STATUS_FORBIDDEN,         // 403
     HTTP_STATUS_NOT_FOUND,         // 404
+    HTTP_STATUS_METHOD_NOT_ALLOWED,// 405
     HTTP_STATUS_PAYLOAD_TOO_LARGE, // 413
     HTTP_STATUS_URI_TOO_LONG,      // 414
     HTTP_STATUS_INTERNAL_SERVER_ERROR,          // 500
@@ -255,30 +310,12 @@ timespec_sub (struct timespec const *a, struct timespec const *b)
 }
 
 
+/* Memory allocation primitives. */
+#include "mem.h"
 /* List definitions. */
 #include "list.h"
-
-/* Locations in the source file */
-struct locus_point
-{
-  char const *filename;
-  int line;
-  int col;
-};
-
-struct locus_range
-{
-  struct locus_point beg, end;
-};
-
-typedef struct workdir
-{
-  DLIST_ENTRY (workdir) link;
-  int refcount;
-  int fd;
-  char name[1];
-} WORKDIR;
-
+/* Configuration parser */
+#include "cfgparser.h"
 
 /* Header types */
 enum
@@ -312,12 +349,26 @@ struct http_header
   DLIST_ENTRY (http_header) link;
 };
 
+static inline char const *
+http_header_name_ptr (struct http_header *hdr)
+{
+  return hdr->header + hdr->name_start;
+}
+
+static inline size_t
+http_header_name_len (struct http_header *hdr)
+{
+  return hdr->name_end - hdr->name_start;
+}
+
 typedef DLIST_HEAD(,http_header) HTTP_HEADER_LIST;
 
+/* Append modes: what to do if the header with that name already exist. */
 enum
   {
-    H_DROP,
-    H_REPLACE
+    H_KEEP,     /* Keep old header, discard changes. */
+    H_REPLACE,  /* Replace old header with the new one. */
+    H_APPEND    /* Append to the value (assume #(values)) */
   };
 
 int http_header_list_append (HTTP_HEADER_LIST *head, char *text, int replace);
@@ -371,10 +422,22 @@ typedef SLIST_HEAD (,acl) ACL_HEAD;
 
 int acl_match (ACL *acl, struct sockaddr *sa);
 
+enum job_ctl
+  {
+    job_ctl_run,
+    job_ctl_cancel
+  };
+typedef void (*JOB_FUNC) (enum job_ctl, void *, const struct timespec *);
+typedef unsigned long JOB_ID;
+
+JOB_ID job_enqueue (struct timespec const *ts, JOB_FUNC func, void *data);
+void job_cancel (JOB_ID id);
+int job_get_timestamp (JOB_ID jid, struct timespec *ts);
+
 /* matcher chain */
 typedef struct _matcher
 {
-  regex_t pat;		/* pattern to match the request/header against */
+  GENPAT pat;		/* pattern to match the request/header against */
   SLIST_ENTRY (_matcher) next;
 } MATCHER;
 
@@ -395,7 +458,8 @@ typedef enum
 
 typedef enum
   {
-    BE_BACKEND,
+    BE_REGULAR,
+    BE_MATRIX,
     BE_REDIRECT,
     BE_ACME,
     BE_CONTROL,
@@ -405,6 +469,44 @@ typedef enum
   }
   BACKEND_TYPE;
 
+enum backend_resolve_mode
+  {
+    bres_immediate,
+    bres_first,
+    bres_all,
+    bres_srv
+  };
+
+char const *resolve_mode_str (int mode);
+
+typedef struct backend_table *BACKEND_TABLE;
+
+struct be_matrix
+{
+  char *hostname;       /* Hostname or IP address. */
+  int port;             /* Port number (network order). */
+  int family;           /* Address family for resolving hostname. */
+  int resolve_mode;     /* Mode for resolving hostname. */
+  unsigned retry_interval; /* Retry interval for failed queries. */
+  int ignore_srv_weight;   /* Ignore weight field in SRV RR. */
+  unsigned override_ttl;   /* Use this TTL instead of one returned from DNS. */
+
+  unsigned to;		/* read/write time-out */
+  unsigned conn_to;	/* connection time-out */
+  unsigned ws_to;	/* websocket time-out */
+  SSL_CTX *ctx;		/* CTX for SSL connections */
+  char *servername;     /* SNI */
+
+  BACKEND_TABLE betab;  /* Table of regular backends generated from this
+			   matrix. */
+  JOB_ID jid;           /* ID of the periodic job scheduled to update this
+			   matrix. */
+  int weight;           /* Weight of the backend list where to allocate
+			   regular backends. */
+  struct _backend *parent;  /* Points to matrix backend, if this backend was
+			       dynamically generated. */
+};
+
 struct be_regular
 {
   struct addrinfo addr;	/* IPv4/6 address */
@@ -413,6 +515,10 @@ struct be_regular
   unsigned conn_to;	/* connection time-out */
   unsigned ws_to;	/* websocket time-out */
   SSL_CTX *ctx;		/* CTX for SSL connections */
+  char *servername;     /* SNI */
+
+  struct _backend *parent; /* Points to matrix backend, if this backend was
+			      dynamically generated. */
 };
 
 struct be_redirect
@@ -437,14 +543,24 @@ struct be_error
 typedef struct _backend
 {
   struct _service *service;     /* Back pointer to the owning service */
-  char *locus;                  /* Location in the config file */
+  struct balancer *balancer;    /* Back pointer to the owning backend list. */
+  struct locus_range locus;     /* Location in the config file */
+  char *locus_str;              /* Location formatted as string. */
   BACKEND_TYPE be_type;         /* Backend type */
   int priority;			/* priority */
   int disabled;			/* true if the back-end is disabled */
-  SLIST_ENTRY (_backend) next;
+  DLIST_ENTRY (_backend) link;
+
+  /* Auxiliary fields. */
+  int mark;                     /* If set, this backend is a candidate for
+				   deletion. */
+
 
   /* Statistics */
   pthread_mutex_t mut;		/* mutex for this back-end */
+#ifdef ENABLE_DYNAMIC_BACKENDS
+  unsigned long refcount;       /* reference counter */
+#endif
   double numreq;		/* number of requests seen */
   double avgtime;		/* Avg. time per request */
   double avgsqtime;             /* Avg. squared time per request */
@@ -453,6 +569,7 @@ typedef struct _backend
   union
   {
     struct be_regular reg;
+    struct be_matrix mtx;
     struct be_acme acme;
     struct be_redirect redirect;
     struct be_error error;
@@ -461,19 +578,29 @@ typedef struct _backend
 
 } BACKEND;
 
-typedef SLIST_HEAD (,_backend) BACKEND_HEAD;
+typedef DLIST_HEAD (,_backend) BACKEND_HEAD;
 
 static inline int backend_is_https (BACKEND *be)
 {
-  return be->be_type == BE_BACKEND && be->v.reg.ctx != NULL;
+  return be->be_type == BE_REGULAR && be->v.reg.ctx != NULL;
 }
 
 static inline int backend_is_alive (BACKEND *be)
 {
-  /* Redirects, ACME, and control backends are always alive */
-  return be->be_type != BE_BACKEND || be->v.reg.alive;
+  /* Matrix backends are special, they are never alive. */
+  if (be->be_type == BE_MATRIX)
+    return 0;
+  /* Redirects, ACME, and control backends are always alive. */
+  return be->be_type != BE_REGULAR || be->v.reg.alive;
 }
 
+static inline int backend_is_active (BACKEND *be)
+{
+  return !be->disabled && backend_is_alive (be);
+}
+
+BACKEND *backend_create (BACKEND_TYPE type, int prio, struct locus_range *loc);
+
 typedef struct session
 {
   char *key;
@@ -482,16 +609,13 @@ typedef struct session
   DLIST_ENTRY (session) link;
 } SESSION;
 
-/* maximal session key size */
 #define KEY_SIZE    127
 
-#if OPENSSL_VERSION_NUMBER >= 0x10100000L
-DEFINE_LHASH_OF (SESSION);
-#else
-DECLARE_LHASH_OF (SESSION);
-#endif
-
-typedef LHASH_OF (SESSION) SESSION_HASH;
+#define HT_TYPE SESSION
+#define HT_NAME_FIELD key
+#define HT_NO_FOREACH
+#define HT_NO_HASH_FREE
+#include "ht.h"
 
 typedef struct
 {
@@ -523,10 +647,11 @@ enum service_cond_type
     COND_QUERY, /* Raw query match. */
     COND_QUERY_PARAM, /* Query parameter match */
     COND_HDR,   /* Header match. */
-    COND_HOST,  /* Special case od COND_HDR: matches the value of the
+    COND_HOST,  /* Special case of COND_HDR: matches the value of the
 		   Host: header */
     COND_BASIC_AUTH,  /* Check if request passes basic auth. */
     COND_STRING_MATCH,/* String match. */
+    COND_CLIENT_CERT
   };
 
 typedef struct string_ref
@@ -538,7 +663,7 @@ typedef struct string_ref
 struct string_match
 {
   STRING_REF *string;
-  regex_t re;
+  GENPAT re;
 };
 
 struct user_pass
@@ -565,11 +690,12 @@ typedef struct _service_cond
   union
   {
     ACL *acl;
-    regex_t re;
+    GENPAT re;
     struct bool_service_cond bool;
     struct _service_cond *cond;
     struct string_match sm;  /* COND_QUERY_PARAM and COND_STRING_MATCH */
     struct pass_file pwfile; /* COND_BASIC_AUTH */
+    X509 *x509;              /* COND_CLIENT_CERT */
   };
   SLIST_ENTRY (_service_cond) next;
 } SERVICE_COND;
@@ -633,9 +759,11 @@ typedef struct rewrite_rule
 
 typedef enum
   {
-    BALANCER_RANDOM,
-    BALANCER_IWRR,
-  } BALANCER;
+    BALANCER_ALGO_RANDOM,
+    BALANCER_ALGO_IWRR,
+  } BALANCER_ALGO;
+
+#define PRI_MAX   65535
 
 enum
   {
@@ -643,33 +771,63 @@ enum
     REWRITE_RESPONSE
   };
 
+struct iwrr_balancer
+{
+  int round;
+  int max_pri;                  /* maximum priority */
+  BACKEND *cur;
+};
+
+struct rand_balancer
+{
+  unsigned long sum_pri;	/* sum of priorities of active backends */
+};
+
+typedef struct balancer
+{
+  BALANCER_ALGO algo;
+  int weight;                   /* relative weight among other balancers */
+  unsigned act_num;             /* number of active backends */
+  BACKEND_HEAD backends;
+  DLIST_ENTRY (balancer) link;
+  union
+  {
+    struct iwrr_balancer iwrr;
+    struct rand_balancer rand;
+  };
+} BALANCER;
+
+typedef DLIST_HEAD (,balancer) BALANCER_LIST;
+
 /* service definition */
 typedef struct _service
 {
   char *name;			/* symbolic name */
-  char *locus;                  /* Location in the config file */
+  char *locus_str;              /* Location in the config file, as string. */
   SERVICE_COND cond;
   REWRITE_RULE_HEAD rewrite[2];
-  BACKEND_HEAD backends;
-  BACKEND *emergency;
-  int abs_pri;			/* abs total priority for all back-ends */
-  int tot_pri;			/* total priority for current back-ends */
-  int max_pri;                  /* maximum priority */
-  BALANCER balancer;
-  /* For IWRR balancer */
-  int iwrr_round;
-  BACKEND *iwrr_cur;
+  BALANCER_LIST balancers;
+  BALANCER_ALGO balancer_algo;
   pthread_mutex_t mut;		/* mutex for this service */
   SESS_TYPE sess_type;
   unsigned sess_ttl;		/* session time-to-live */
   char *sess_id;                /* Session anchor ID */
   SESSION_TABLE *sessions;	/* currently active sessions */
   int disabled;			/* true if the service is disabled */
+
   /* Logging */
   char *forwarded_header;       /* "forwarded" header name */
   ACL *trusted_ips;             /* Trusted IP addresses */
   int log_suppress_mask;        /* Suppress HTTP logging for these status
 				   codes.  A bitmask. */
+
+  /* Backend removal */
+  BACKEND_HEAD be_rem_head;     /* List of backends scheduled for removal. */
+  pthread_cond_t be_rem_cond;   /* Condition through which the removal thread
+				   is notified that a backend from the list is
+				   ready for removal.
+				 */
+
   SLIST_ENTRY (_service) next;
 } SERVICE;
 
@@ -706,7 +864,7 @@ int http_log_format_check (int n);
 typedef struct _listener
 {
   char *name;			/* symbolic name */
-  char *locus;                  /* Location in the config file */
+  char *locus_str;              /* Location in the config file, as string. */
   struct addrinfo addr;		/* Socket address */
   int mode;                     /* File mode for AF_UNIX */
   int chowner;                  /* Change to effective owner, for AF_UNIX */
@@ -718,10 +876,10 @@ typedef struct _listener
   REWRITE_RULE_HEAD rewrite[2];
   int verb;			/* allowed HTTP verb group */
   unsigned to;			/* client time-out */
-  int has_pat;			/* was a URL pattern defined? */
-  regex_t url_pat;		/* pattern to match the request URL against */
+  GENPAT url_pat;	/* pattern to match the request URL against */
   char *http_err[HTTP_STATUS_MAX];	/* error messages */
-  CONTENT_LENGTH max_req;	/* max. request size */
+  CONTENT_LENGTH max_req_size;	/* max. request size */
+  unsigned max_uri_length;      /* max. URI length */
   int rewr_loc;			/* rewrite location response */
   int rewr_dest;		/* rewrite destination header */
   int disabled;			/* true if the listener is disabled */
@@ -733,9 +891,12 @@ typedef struct _listener
   SLIST_ENTRY (_listener) next;
 
   /* Used during configuration parsing */
+  char *addr_str;
+  char *port_str;
   int ssl_op_enable;
   int ssl_op_disable;
-  int has_other;
+  int verify;
+  int socket_from;
 } LISTENER;
 
 typedef SLIST_HEAD(,_listener) LISTENER_HEAD;
@@ -744,7 +905,7 @@ struct submatch
 {
   size_t matchn;
   size_t matchmax;
-  regmatch_t *matchv;
+  POUND_REGMATCH *matchv;
   char *subject;
 };
 
@@ -834,25 +995,6 @@ typedef struct _pound_http
 
 typedef SLIST_HEAD(,_pound_http) POUND_HTTP_HEAD;
 
-/* control request stuff */
-typedef enum
-{
-  CTRL_LST,
-  CTRL_EN_LSTN, CTRL_DE_LSTN,
-  CTRL_EN_SVC, CTRL_DE_SVC,
-  CTRL_EN_BE, CTRL_DE_BE,
-  CTRL_ADD_SESS, CTRL_DEL_SESS
-} CTRL_CODE;
-
-typedef struct
-{
-  CTRL_CODE cmd;
-  int listener;
-  int service;
-  int backend;
-  char key[KEY_SIZE + 1];
-} CTRL_CMD;
-
 void save_forwarded_header (POUND_HTTP *phttp);
 void http_log (POUND_HTTP *phttp);
 
@@ -888,8 +1030,32 @@ SERVICE *get_service (POUND_HTTP *);
 /* Find the right back-end for a request */
 BACKEND *get_backend (POUND_HTTP *phttp);
 
+#ifdef ENABLE_DYNAMIC_BACKENDS
+void backend_ref (BACKEND *be);
+void backend_unref (BACKEND *be);
+static inline void
+backend_refcount_init (BACKEND *be)
+{
+  be->refcount = 1;
+}
+static inline int backend_referenced (BACKEND *be)
+{
+  return be->refcount > 1;
+}
+#else
+# define backend_ref(be)
+# define backend_unref(be)
+# define backend_refcount_init(be)
+# define backend_referenced(be) 1
+#endif
+
+void backend_matrix_to_regular (struct be_matrix *mtx, struct addrinfo *addr,
+				struct be_regular *reg);
+void backend_matrix_init (BACKEND *be);
+void backend_matrix_disable (BACKEND *be, int disable_mode);
+
 /* Search for a host name, return the addrinfo for it */
-int get_host (char *const, struct addrinfo *, int);
+int get_host (char const *, struct addrinfo *, int);
 
 /*
  * Find if a redirect needs rewriting
@@ -910,14 +1076,50 @@ void upd_session (SERVICE *, HTTP_HEADER_LIST *, BACKEND *);
 
 /*
  * mark a backend host as dead;
- * do nothing if no resurection code is active
+ * do nothing if no resurrection code is active
  */
 void kill_be (SERVICE *, BACKEND *, const int);
 
-/*
- * Update the number of requests and time to answer for a given back-end
- */
-void upd_be (SERVICE * const svc, BACKEND * const be, const double);
+void service_session_remove_by_backend (SERVICE *svc, BACKEND *be);
+void service_recompute_pri (SERVICE *svc,
+			    BALANCER *bl,
+			    void (*cb) (BACKEND *, void *),
+			    void *data);
+void service_recompute_pri_unlocked (SERVICE *svc,
+				     void (*cb) (BACKEND *, void *),
+				     void *data);
+void balancer_recompute_pri_unlocked (BALANCER *bl,
+					  void (*cb) (BACKEND *, void *),
+					  void *data);
+
+static inline void balancer_add_backend (BALANCER *bl, BACKEND *be)
+{
+  be->balancer = bl;
+  DLIST_INSERT_TAIL (&bl->backends, be, link);
+}
+
+static inline void balancer_remove_backend (BALANCER *bl, BACKEND *be)
+{
+  be->balancer = NULL;
+  DLIST_REMOVE (&bl->backends, be, link);
+}
+
+BALANCER *balancer_list_alloc (BALANCER_LIST *ml);
+BALANCER *balancer_list_get (BALANCER_LIST *ml, int n, BALANCER_ALGO algo);
+
+#define BALANCER_WEIGTH_MAX  65535
+
+static inline BALANCER *
+balancer_list_get_normal (BALANCER_LIST *ml)
+{
+  return balancer_list_get (ml, 0, BALANCER_ALGO_RANDOM);
+}
+
+static inline BALANCER *
+balancer_list_get_emerg (BALANCER_LIST *ml)
+{
+  return balancer_list_get (ml, BALANCER_WEIGTH_MAX, BALANCER_ALGO_RANDOM);
+}
 
 /*
  * Non-blocking version of connect(2). Does the same as connect(2) but
@@ -954,81 +1156,8 @@ void *thr_timer (void *);
 void POUND_SSL_CTX_init (SSL_CTX *ctx);
 int set_ECDHCurve (char *name);
 
-void *mem2nrealloc (void *p, size_t *pn, size_t s);
-void xnomem (void);
-void *xmalloc (size_t s);
-void *xcalloc (size_t nmemb, size_t size);
-#define xzalloc(s) xcalloc(1, s)
-#define XZALLOC(v) (v = xzalloc (sizeof ((v)[0])))
-
-void *xrealloc (void *p, size_t s);
-void *x2nrealloc (void *p, size_t *pn, size_t s);
-char *xstrdup (char const *s);
-char *xstrndup (const char *s, size_t n);
-
-struct stringbuf
-{
-  char *base;                     /* Buffer storage. */
-  size_t size;                    /* Size of buf. */
-  size_t len;                     /* Actually used length in buf. */
-  void (*nomem) (void);           /* Out of memory handler. */
-  int err;                        /* Error indicator */
-};
-
-void stringbuf_init (struct stringbuf *sb, void (*nomem) (void));
-void stringbuf_reset (struct stringbuf *sb);
-char *stringbuf_finish (struct stringbuf *sb);
-void stringbuf_free (struct stringbuf *sb);
-int stringbuf_add (struct stringbuf *sb, char const *str, size_t len);
-int stringbuf_add_char (struct stringbuf *sb, int c);
-int stringbuf_add_string (struct stringbuf *sb, char const *str);
-int stringbuf_vprintf (struct stringbuf *sb, char const *fmt, va_list ap);
-int stringbuf_printf (struct stringbuf *sb, char const *fmt, ...)
-  ATTR_PRINTFLIKE(2,3);
-char *stringbuf_set (struct stringbuf *sb, int c, size_t n);
-int stringbuf_strftime (struct stringbuf *sb, char const *fmt,
-			const struct tm *tm);
-
-static inline int
-stringbuf_err (struct stringbuf *sb)
-{
-  return sb->err;
-}
-
-static inline char *stringbuf_value (struct stringbuf *sb)
-{
-  return sb->base;
-}
-
-static inline size_t stringbuf_len (struct stringbuf *sb)
-{
-  return sb->len;
-}
-
-extern void xnomem (void);
-extern void lognomem (void);
-
-static inline void xstringbuf_init (struct stringbuf *sb)
-{
-  stringbuf_init (sb, xnomem);
-}
-
-static inline void stringbuf_init_log (struct stringbuf *sb)
-{
-  stringbuf_init (sb, lognomem);
-}
-
-void job_enqueue_unlocked (struct timespec const *ts, void (*func) (void *), void *data);
-void job_enqueue_after_unlocked (unsigned t, void (*func) (void *), void *data);
-
-void job_enqueue (struct timespec const *ts, void (*func) (void *), void *data);
-void job_enqueue_after (unsigned t, void (*func) (void *), void *data);
-
-void job_rearm_unlocked (struct timespec *ts, void (*func) (void *), void *data);
-void job_rearm (struct timespec *ts, void (*func) (void *), void *data);
-
 char const *sess_type_to_str (int type);
-int control_response (POUND_HTTP *arg);
+int control_response_basic (POUND_HTTP *arg);
 void pound_atexit (void (*func) (void *), void *arg);
 int unlink_at_exit (char const *file_name);
 
@@ -1079,8 +1208,10 @@ void template_run (TEMPLATE tmpl, struct json_value *val, FILE *outfile);
 char const *template_strerror (int ec);
 void template_free (TEMPLATE tmpl);
 
-void errormsg (int ex, int ec, char const *fmt, ...);
-void json_error (struct json_value *val, char const *fmt, ...);
+void errormsg (int ex, int ec, char const *fmt, ...)
+  ATTR_PRINTFLIKE(3,4);
+void json_error (struct json_value *val, char const *fmt, ...)
+  ATTR_PRINTFLIKE(2,3);
 
 int http_status_to_pound (int status);
 int pound_to_http_status (int err);
@@ -1093,6 +1224,7 @@ int match_cond (SERVICE_COND *cond, POUND_HTTP *phttp,
 		struct http_request *req);
 
 struct http_header *http_header_list_locate_name (HTTP_HEADER_LIST *head, char const *name, size_t len);
+struct http_header *http_header_list_next (struct http_header *hdr);
 char *http_header_get_value (struct http_header *hdr);
 
 /*
@@ -1117,10 +1249,13 @@ int http_request_get_basic_auth (struct http_request *req,
 				 char **u_name, char **u_pass);
 
 void service_lb_init (SERVICE *svc);
+void service_lb_reset (SERVICE *svc, BACKEND *be);
 
 FILE *fopen_wd (WORKDIR *wd, const char *filename);
+FILE *fopen_include (const char *filename);
 void fopen_error (int pri, int ec, WORKDIR *wd, const char *filename,
 		  struct locus_range *loc);
+char *filename_resolve (const char *filename);
 
 typedef int (*LISTENER_ITERATOR) (LISTENER *, void *);
 int foreach_listener (LISTENER_ITERATOR itr, void *data);
@@ -1132,3 +1267,6 @@ typedef int (*BACKEND_ITERATOR) (BACKEND *, void *);
 int foreach_backend (BACKEND_ITERATOR itr, void *data);
 
 int basic_auth (struct pass_file *pwf, struct http_request *req);
+
+void combinable_header_add (char const *name);
+int is_combinable_header (struct http_header *hdr);
