@@ -1,5 +1,5 @@
 # This file is part of pound testsuite
-# Copyright (C) 2018-2023 Sergey Poznyakoff
+# Copyright (C) 2018-2025 Sergey Poznyakoff
 #
 # Pound is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -23,6 +23,7 @@ use HTTP::Tiny;
 use POSIX qw(:sys_wait_h);
 use Cwd qw(abs_path);
 use File::Spec;
+use Socket;
 
 my $config = 'pound.cfi:pound.cfg';
 my @preproc_files;
@@ -38,15 +39,21 @@ my $backends = ListenerList->new(0);
 my $pound_pid;
 my $startup_timeout = 2;
 my $statistics;
+my $fakedns;
+my $valgrind;
+my @source_addr;
 
 use constant {
     EX_SUCCESS => 0,
     EX_FAILURE => 1,
     EX_ERROR => 2,
     EX_USAGE => 3,
-    EX_SKIP => 77
+    EX_SKIP => 77,
+    EX_EXEC => 127
 };
 
+my $nameserver;
+
 ## Early checks
 ## ------------
 
@@ -64,10 +71,16 @@ if ($@) {
     exit(EX_SKIP);
 }
 
+eval "require JSON";
+if ($@) {
+    print STDERR "required module JSON not present\n";
+    exit(::EX_SKIP);
+}
+
 sub cleanup {
     if ($pound_pid) {
 	if ($verbose) {
-	    print "Stopping pound ($pound_pid)\n";
+	    print "$$ Stopping pound ($pound_pid)\n";
 	}
 	kill 'HUP', $pound_pid;
     }
@@ -77,16 +90,11 @@ sub cleanup {
 ## -----------------------------------
 
 my %status_codes;
+share(%status_codes);
 
 $SIG{QUIT} = $SIG{HUP} = $SIG{TERM} = $SIG{INT} = \&cleanup;
-sub sigchild {
-    my $pid = waitpid(-1, 0);
-    if ($pid != $pound_pid) {
-	$status_codes{$pid} = $?;
-	$SIG{CHLD} = \&sigchild;
-	return;
-    }
-    $pound_pid = 0;
+
+sub handle_pound_status {
     if (WIFEXITED($?)) {
 	if (WEXITSTATUS($?)) {
 	    print STDERR "pound terminated with code " . WEXITSTATUS($?) . "\n";
@@ -94,22 +102,77 @@ sub sigchild {
 	    if ($verbose) {
 		print "pound finished\n";
 	    }
-	    $SIG{CHLD} = \&sigchild;
-	    return;
+	    return 1;
 	}
     } elsif (WIFSIGNALED($?)) {
 	print STDERR "pound terminated on signal " . WTERMSIG($?) . "\n";
     } else {
 	print STDERR "pound terminated with unrecognized status " . $? . "\n";
     }
+
+    return 0;
+}
+
+sub sigchild {
+    my $pid = waitpid(-1, 0);
+    if ($pid == -1) {
+	return
+    }
+    if (!defined($pound_pid) || $pid != $pound_pid) {
+	$status_codes{$pid} = $?;
+	$SIG{CHLD} = \&sigchild;
+	return;
+    }
+    $pound_pid = 0;
+    if (handle_pound_status()) {
+	$SIG{CHLD} = \&sigchild;
+	return;
+    }
     exit(EX_ERROR);
 };
 $SIG{CHLD} = \&sigchild;
+
+sub fakedns_self_check {
+    my $params = 'ns1.example.org. rname@example.org. 42 7200 3600 1209600 3600';
 
-END {
-    cleanup;
+    $nameserver->ZoneUpdate(
+	'$ORIGIN example.org.',
+	'@ IN SOA '.$params
+    );
+
+    local $ENV{LD_PRELOAD} = $fakedns;
+    if (open(my $fh, '-|', 'getsoa', 'example.org', '192.0.2.24')) {
+	chomp(my $s = <$fh>);
+	if ($s ne $params) {
+	    print STDERR "Self-check: got $s\n";
+	    return;
+	}
+	if (<$fh>) {
+	    print STDERR "Self-check: garbage after response: $_\n";
+	    return;
+	}
+	close $fh;
+	return 1;
+    }
 }
-
+
+sub assert_source_ip {
+    my ($ip) = @_;
+    socket(my $sock, PF_INET, SOCK_STREAM, 0)
+	or do {
+	    warn "socket: $!";
+	    exit 77;
+    };
+    setsockopt($sock, SOL_SOCKET, SO_REUSEADDR, 1);
+    unless (bind($sock, sockaddr_in(0, inet_aton($ip)))) {
+	# FIXME: To be strict, this should check for $!{EADDRNOTAVAIL} and
+	# exit 77 if so.
+	warn "bind($ip): $!";
+	exit(77);
+    }
+    close($sock)
+}
+
 ## Start the program
 ## -----------------
 sub usage_error {
@@ -125,12 +188,43 @@ GetOptions('config|f=s' => \$config,
 	   'transcript|x=s' => \$transcript_file,
 	   'statistics|s' => \$statistics,
 	   'startup-timeout|t=n' => \$startup_timeout,
-	   'include-dir=s' => \$include_dir)
+	   'include-dir=s' => \$include_dir,
+	   'fakedns=s' => \$fakedns,
+           'valgrind=s' => \$valgrind,
+	   'source-address|a=s@' => sub {
+	       my (undef, $ip) = @_;
+	       assert_source_ip($ip);
+	       push @source_addr, $ip;
+	   })
     or exit(EX_USAGE);
 
 my $script_file = shift @ARGV or usage_error "required parameter missing";
 usage_error "too many arguments\n" if @ARGV;
 
+if ($fakedns) {
+    eval "require PoundNS";
+    if ($@) {
+	print STDERR "PoundNS: $@\n";
+	exit(EX_SKIP);
+    }
+    unless (-e $fakedns) {
+	print STDERR "$fakedns: file not found\n";
+	exit(EX_SKIP);
+    }
+
+    if (!File::Spec->file_name_is_absolute($fakedns)) {
+	$fakedns = File::Spec->rel2abs($fakedns);
+    }
+
+    $nameserver = PoundNS->new or die "can't create nameserver";
+    ($ENV{FAKEDNS_UDP_PORT}, $ENV{FAKEDNS_TCP_PORT}) = $nameserver->start_server();
+
+    unless (fakedns_self_check()) {
+	print STDERR "Self-check failed\n";
+	exit(EX_SKIP);
+    }
+}
+
 foreach my $file (@preproc_files) {
     if ($file =~ m{(?<src>.+):(?<dst>.+)}) {
 	preproc($+{src}, $+{dst});
@@ -152,6 +246,8 @@ if ($include_dir) {
     $config = abs_path($config)
 }
 
+$ENV{POUNDCTL_CONF} = "";
+
 # Start HTTP listeners
 $backends->read_and_process;
 
@@ -159,7 +255,8 @@ $backends->read_and_process;
 runner();
 
 # Parse and run the script file
-my $ps = PoundScript->new($script_file, $transcript_file);
+my $ps = PoundScript->new($script_file, $transcript_file, $nameserver,
+			  \@source_addr);
 $ps->parse;
 
 # Terminate
@@ -170,6 +267,17 @@ if ($statistics) {
 } elsif ($ps->failures) {
     print STDERR $ps->failures . " from " . $ps->tests . " tests failed\n";
 }
+
+if (defined($pound_pid) && $pound_pid > 0) {
+    $SIG{CHLD} = sub { };
+    cleanup;
+    if (waitpid($pound_pid, 0) == $pound_pid) {
+	if (handle_pound_status() == 0) {
+	    exit(EX_ERROR);
+	}
+    }
+}
+
 exit ($ps->failures ? EX_FAILURE : EX_SUCCESS);
 
 ## Configuration file processing
@@ -218,7 +326,10 @@ sub preproc {
 	ST_LISTENER => 1,
 	ST_SERVICE => 2,
 	ST_BACKEND => 3,
-	ST_SESSION => 4
+	ST_BACKEND_STAT => 4,
+	ST_BACKEND_DYN => 5,
+	ST_SESSION => 6,
+	ST_SECTION => 7
     };
     my @state;
     unshift @state, $initstate // ST_INIT;
@@ -226,14 +337,26 @@ sub preproc {
 	print $out <<EOT;
 # Initial settings by $0
 Daemon 0
+Control "pound.ctl"
 LogFacility -
 EOT
     ;
 	if ($log_level >= 0) {
 	    print $out "LogLevel $log_level\n";
 	}
+	if ($fakedns) {
+	    print $out <<EOT;
+Resolver
+    ConfigText
+	nameserver 192.0.2.24
+    End
+    RetryInterval 10
+End
+EOT
+    ;
+	}
     }
-    my $be;
+    my $be_loc;
     while (<$in>) {
 	chomp;
 	if (/^\s*(?:#.)?$/) {
@@ -263,19 +386,40 @@ EOT
 	} elsif (/^\s*Session/i) {
 	    unshift @state, ST_SESSION;
 	} elsif (/^\s*(Backend|Emergency)/i) {
+	    $be_loc = "$infile:$.";
 	    unshift @state, ST_BACKEND;
-	    $be = $backends->create("$infile:$.");
-	    if ($verbose) {
-		print "$infile:$.: Backend ".$be->ident . ": " . $be->address."\n";
+	} elsif (/^\s*((Match)|(Rewrite)|(TrustedIP)|(ACL)|(CombineHeaders))\b/i) {
+	    unshift @state, ST_SECTION
+	} elsif (/^(\s*)End/i) {
+	    if ($state[0] == ST_BACKEND) {
+		my $be = $backends->create($be_loc);
+		if ($verbose) {
+		    print "$infile:$.: Backend ".$be->ident . ": " . $be->address."\n";
+		}
+		print $out "# Inserted by $0\n";
+		print $out "$1\tAddress ".$be->host."\n";
+		print $out "$1\tPort ".$be->port."\n";
 	    }
-	} elsif (/^\s*End/i) {
 	    shift @state
 	} elsif ($state[0] == ST_BACKEND) {
-	    if (/^(\s*Address)/i) {
-		$_ = $1 . ' ' . $be->host;
-	    } elsif (/^(\s*Port)/i) {
-		$_ = $1 . ' ' . $be->port;
+	    shift @state;
+	    if (/^\s*Resolve/) {
+		unshift @state, ST_BACKEND_DYN;
+	    } else {
+		unshift @state, ST_BACKEND_STAT;
+		my $be = $backends->create($be_loc);
+		if ($verbose) {
+		    print "$infile:$.: Backend ".$be->ident . ": " . $be->address."\n";
+		}
+		/^(\s*)([\S]+)/;
+		my ($indent, $kw) = ($1, $2);
+		print $out "# Inserted by $0\n";
+		print $out "${indent}Address ".$be->host."\n";
+		print $out "${indent}Port ".$be->port."\n";
+		next if ($kw =~ /^Address/i || $kw =~ /^Port/i);
 	    }
+	} elsif ($state[0] == ST_BACKEND_STAT) {
+	    next if (/^\s*Address/i || /^\s*Port/i);
 	} elsif ($state[0] == ST_LISTENER) {
 	    if (/^\s*(Address|Port|SocketFrom)/i) {
 		    $_ = "# Commented out: $_";
@@ -290,7 +434,7 @@ EOT
 		    my $file;
 		    while ($hostline =~ m/^\s*(-\S+)(.*)/) {
 			push @opts, $1;
-			if ($1 eq '-re' || $1 eq '-beg' || $1 eq '-end') {
+			if ($1 eq '-re' || $1 eq '-beg') {
 			    $exact = 0;
 			} elsif ($1 eq '-file') {
 			    $file = 1;
@@ -347,8 +491,19 @@ sub runner {
     }
     open(STDOUT, '>', $log_file);
     open(STDERR, ">&STDOUT");
-    exec 'pound', '-p', $pid_file, '-f', $config, '-v', '-W', 'no-dns', '-W',
-	  $include_dir ? "include-dir=$include_dir" : 'no-include-dir';
+    my @cmd = (
+	'pound', '-p', $pid_file, '-f', $config, '-v',
+	 '-W', $include_dir ? "include-dir=$include_dir" : 'no-include-dir'
+    );
+    if ($valgrind) {
+	unshift @cmd, 'valgrind', '--log-file=' . $valgrind;
+    }
+    if ($fakedns) {
+	$ENV{LD_PRELOAD} = $fakedns;
+    } else {
+	push @cmd, '-W', 'no-dns';
+    }
+    exec @cmd or exit(EX_EXEC);
 }
 
 package PoundScript;
@@ -362,8 +517,8 @@ use IPC::Open3;
 use Symbol 'gensym';
 
 sub new {
-    my ($class, $file, $xscript) = @_;
-    my $self = bless { tests => 0, failures => 0 }, $class;
+    my ($class, $file, $xscript, $ns, $srcaddr) = @_;
+    my $self = bless { tests => 0, failures => 0, ns => $ns }, $class;
     if ($file ne '-') {
 	open(my $fh, '<', $file)
 	    or croak "can't open script $file: $!";
@@ -378,14 +533,59 @@ sub new {
 	    or croak "can't create transcript file $xscript: $!";
     }
     $self->{server} = 0;
-    $self->{http} = HTTP::Tiny->new(max_redirect => 0);
+    $self->{source_addr} = '127.0.0.1';
+    if ($srcaddr) {
+	unshift @$srcaddr, $self->{source_addr}
+    } else {
+	$srcaddr = [ $self->{source_addr} ]
+    }
+    foreach my $addr (@$srcaddr) {
+	$self->{http}{$addr} = HTTP::Tiny->new(
+			  max_redirect => 0,
+			  verify_SSL => 0,
+			  local_address => $addr
+	    );
+    }
     return $self;
 }
 
 sub tests { shift->{tests} }
 sub failures { shift->{failures} }
 
-sub http { shift->{http} }
+sub http {
+    my ($self) = @_;
+    return $self->{http}{$self->{source_addr}}
+}
+
+sub source_address {
+    my ($self, $addr) = @_;
+    croak "unknown source address" unless exists $self->{http}{$addr};
+    $self->{source_addr} = $addr
+}
+
+sub transcript {
+    my ($self, $text, %opt) = @_;
+    if (my $fh = $self->{xscript}) {
+	if (my $locus = delete $opt{locus}) {
+	    print $fh "$self->{filename}:$locus->{BEG}-$locus->{END}";
+	} else {
+	    print $fh "$self->{filename}:$self->{line}";
+	}
+	print $fh ": ";
+	print $fh $text;
+	print $fh "\n";
+    }
+}
+
+sub transcript_ml {
+    my $self = shift;
+    if (my $fh = $self->{xscript}) {
+	print $fh join("\n", @_)."\n";
+    }
+}
+
+sub transcript_ok { shift->transcript("OK", @_) }
+sub transcript_fail { shift->transcript("FAIL", @_) }
 
 sub send_and_expect {
     my ($self, $lst, $host, $stats) = @_;
@@ -400,19 +600,25 @@ sub send_and_expect {
 	$options{headers} = $self->{REQ}{HEADERS};
     }
     if (exists($self->{REQ}{BODY})) {
-	$options{content} = $self->{REQ}{BODY};
+	my $body = $self->{REQ}{BODY};
+	if ($body =~ /@@/) {
+	    my @chunks = split /@@/, $body;
+	    $options{content} = sub {
+		return shift(@chunks)
+	    }
+	} else {
+	    $options{content} = $body;
+	}
     }
     if ($verbose) {
 	print "URL $self->{REQ}{METHOD} $url\n";
     }
-    if (my $fh = $self->{xscript}) {
-	print $fh "$self->{filename}:$self->{REQ}{BEG}-$self->{REQ}{END}: sending $self->{REQ}{METHOD} $url to server $self->{server}\n";
-    }
+
+    $self->transcript("sending $self->{REQ}{METHOD} $url to server $self->{server}",
+		      locus => $self->{REQ});
+
     my $response = $self->http->request($self->{REQ}{METHOD}, $url, \%options);
-    if (my $fh = $self->{xscript}) {
-	print $fh "$self->{filename}:$self->{REQ}{BEG}-$self->{REQ}{END}: got:\n";
-	print $fh Dumper([$response]);
-    }
+    $self->transcript("got:\n".Dumper([$response]), locus => $self->{REQ});
 
     $self->{tests}++;
     my $ok = $self->assert($response);
@@ -423,10 +629,7 @@ sub send_and_expect {
     } else {
 	$self->{failures}++;
     }
-    if (my $fh = $self->{xscript}) {
-	print $fh "$self->{filename}:$self->{EXP}{BEG}-$self->{EXP}{END}: " .
-		   ($ok ? "OK" : "FAIL")."\n";
-    }
+    $self->transcript($ok ? "OK" : "FAIL", locus => $self->{EXP});
     return $ok
 }
 
@@ -461,9 +664,8 @@ sub send {
 	    if (exists($self->{stats}{$k})) {
 		unless ($self->check_expect($s->${ \$k }, $self->{stats}{$k})) {
 		    $ok = 0;
-		    if (my $fh = $self->{xscript}) {
-			print $fh "$self->{filename}:$self->{EXP}{BEG}-$self->{EXP}{END}: $k: expected value mismatch\n";
-		    }
+		    $self->transcript("$k: expected value mismatch",
+				      locus => $self->{EXP});
 		}
 	    }
 	}
@@ -475,21 +677,20 @@ sub send {
 		unless ($self->check_expect($s->sample($i),
 					    $self->{stats}{expect}{$i})) {
 		    $ok = 0;
-		    if (my $fh = $self->{xscript}) {
-			print $fh "$self->{filename}:$self->{EXP}{BEG}-$self->{EXP}{END}: element $i: expected value mismatch\n";
-		    }
+		    $self->transcript("element $i: expected value mismatch",
+				      locus => $self->{EXP});
 		}
 	    }
 	}
 
 	$self->{failures}++ unless $ok;
 
-	if (my $fh = $self->{xscript}) {
-	    printf $fh "min=%f, avg=%f, max=%f, stddev=%f\n", $s->min, $s->avg, $s->max, $s->stddev;
+	if ($self->{xscript}) {
+	    $self->transcript(sprintf("min=%f, avg=%f, max=%f, stddev=%f",
+				      $s->min, $s->avg, $s->max, $s->stddev));
 	    local $Data::Dumper::Sortkeys=1;
-	    print $fh Dumper([$s->samples]);
-	    print $fh "$self->{filename}:$self->{EXP}{BEG}-$self->{EXP}{END}: " .
-		($ok ? "OK" : "FAIL")."\n";
+	    $self->transcript(Dumper([$s->samples]));
+	    $self->transcript($ok ? "OK" : "FAIL", locus => $self->{EXP});
 	}
     } else {
 	$self->send_and_expect($lst, $host);
@@ -541,6 +742,8 @@ sub assert {
 	if (exists($self->{EXP}{BODY}) &&
 	    $self->{EXP}{BODY} ne $response->{content}) {
 	    print STDERR "$self->{filename}:$self->{EXP}{BEG}-$self->{EXP}{END}: response content mismatch\n";
+	    print STDERR "EXP '".$self->{EXP}{BODY}."'\n";
+	    print STDERR "GOT '".$response->{content}."'\n";
 	}
     }
     return 1
@@ -549,8 +752,15 @@ sub assert {
 
 sub parse {
     my $self = shift;
-    while (!$self->{eof}) {
-	$self->parse_req;
+    eval {
+	while (!$self->{eof}) {
+	    $self->parse_req;
+	}
+    };
+    if ($@) {
+	print STDERR "$@";
+	print STDERR "Abnormal termination\n";
+	$self->{failures}++;
     }
 }
 
@@ -576,6 +786,33 @@ sub parse_req {
 	    $self->{server} = $1;
 	    next;
 	}
+	if (/^source\s+(.+)/) {
+	    $self->source_address($1);
+	    next;
+	}
+	if (/^mkbackend\s+(127(\.\d+){3})/) {
+	    $backends->create("$self->{filename}:$self->{line}", 'HTTP', $1);
+	    next;
+	}
+	if (/^sleep\s+(\d+)/) {
+	    sleep $1;
+	    next;
+	}
+	if (/^echo\s+(.+)/) {
+	    print STDERR "# ".$self->expandvars($1)."\n";
+	    next;
+	}
+
+	if (/^zonefile/) {
+	    $self->parse_zonefile;
+	    next;
+	}
+
+	if (m{^backends\s+(\d+)\s+(\d+)(?:\s+(.+))?}) {
+	    $self->parse_control_backends($1, $2);
+	    next;
+	}
+
 	if (s/^stats//) {
 	    foreach my $k (qw(samples min max avg stddev index)) {
 		if (s{\s+$k = (?:
@@ -652,6 +889,78 @@ sub parse_req {
     $self->syntax_error("unexpected end of file") if $self->{REQ}{BEG};
 }
 
+sub parse_zonefile {
+    my $self = shift;
+    my $fh = $self->{fh};
+
+    my @zonetext;
+    while (<$fh>) {
+	$self->{line}++;
+	chomp;
+	last if /^end$/;
+	push @zonetext, $self->expandvars($_)
+    }
+
+    $self->{ns}->ZoneUpdate(@zonetext);
+}
+
+sub jsoncmp {
+    my ($a, $b) = @_;
+
+    my $rtype = ref($a);
+    return 0 if $rtype ne ref($b);
+    if ($rtype eq '') {
+	return $a eq $b;
+    } elsif ($rtype eq 'ARRAY') {
+	return 0 if @{$a} != @{$b};
+	for (my $i = 0; $i <= $#{$a}; $i++) {
+	    return 0 unless jsoncmp($a->[$i], $b->[$i]);
+	}
+    } elsif ($rtype eq 'HASH') {
+	foreach my $k (keys %$a) {
+	    return 0 unless jsoncmp($a->{$k}, $b->{$k});
+	}
+    } else {
+	return 0
+    }
+    return 1
+}
+
+sub parse_control_backends {
+    my ($self, $ls, $sv) = @_;
+
+    if ($verbose) {
+	print "querying control interface\n";
+    }
+
+    $self->transcript("querying control interface");
+
+    my $fh = $self->{fh};
+    $self->{tests}++;
+    my @exp;
+    while (<$fh>) {
+	$self->{line}++;
+	chomp;
+	last if /^end$/;
+	push @exp, $self->expandvars($_)
+    }
+
+    my $ctl = PoundControl->new();
+    my $belist = $ctl->backends($ls, $sv, sort => 1);
+    my $json = JSON->new->boolean_values(0,1);
+    my $exp = $json->decode(join(' ', @exp));
+
+    if (jsoncmp($exp, $belist)) {
+	$self->transcript_ok;
+    } else {
+	$self->{failures}++;
+	$self->transcript_ml("backend listings don't match",
+			     "exp: " . $json->canonical->pretty->encode($exp),
+			     "got: " . $json->canonical->pretty->encode($belist));
+	$self->transcript_fail;
+    }
+}
+
 sub parse_headers {
     my $self = shift;
     my $fh = $self->{fh};
@@ -676,7 +985,16 @@ sub parse_headers {
 	}
 
 	if (/^(?<name>[A-Za-z][A-Za-z0-9_-]*):\s*(?<value>.*)$/) {
-	    $self->{REQ}{HEADERS}{lc($+{name})} = $self->expandvars($+{value});
+	    my $name = lc($+{name});
+	    if (exists($self->{REQ}{HEADERS}{$name})) {
+		my $val = $self->{REQ}{HEADERS}{$name};
+		if (ref($val) ne 'ARRAY') {
+		    $self->{REQ}{HEADERS}{$name} = [$val];
+		}
+		push @{$self->{REQ}{HEADERS}{$name}}, $self->expandvars($+{value})
+	    } else {
+		$self->{REQ}{HEADERS}{lc($+{name})} = $self->expandvars($+{value});
+	    }
 	} else {
 	    $self->syntax_error;
 	    return;
@@ -783,7 +1101,7 @@ sub runcom {
     my @ready;
     my %data = ( $child_stdout => '', $child_stderr => '' );
 
-    while (!defined($status_codes{$pid}) && (@ready = $sel->can_read)) {
+    while (@ready = $sel->can_read) {
 	foreach my $fh (@ready) {
 	    my $data;
 	    while (1) {
@@ -812,35 +1130,33 @@ sub runcom {
 	$code = $status_codes{$pid} >> 8;
     }
 
-    if (my $fh = $self->{xscript}) {
-	print $fh "Command: " . $self->{RUNCOM}{command} . "\n";
-	print $fh "Status code: $code\n";
-	print $fh "Stdout:\n";
-	print $fh $data{$child_stdout};
-	print $fh "\nEnd\n";
-	print $fh "Stderr:\n";
-	print $fh $data{$child_stderr};
-	print $fh "\nEnd\n";
-    }
+    $self->transcript_ml(
+	"Command: " . $self->{RUNCOM}{command},
+	"Status code: $code",
+	"Stdout:",
+	$data{$child_stdout},
+	"End",
+	"Stderr:",
+	$data{$child_stderr},
+	"End"
+    );
 
     $self->{tests}++;
+    my $ok = 1;
+
     if (exists($self->{RUNCOM}{status}) && $code != $self->{RUNCOM}{status}) {
 	$self->{failures}++;
-	if (my $fh = $self->{xscript}) {
-	    print $fh "$self->{filename}:$self->{RUNCOM}{BEG}-$self->{RUNCOM}{END}: exit code differs\n";
-	    return 0;
-	}
+	$self->transcript("exit code differs", locus => $self->{RUNCOM});
+	$ok = 0;
     }
 
     if (exists($self->{RUNCOM}{stdout})) {
 	my $s = join("\n", @{$self->{RUNCOM}{stdout}});
 	if ($data{$child_stdout} !~ m{$s}ms) {
 	    $self->{failures}++;
-	    if (my $fh = $self->{xscript}) {
-		print $fh "$self->{filename}:$self->{RUNCOM}{BEG}-$self->{RUNCOM}{END}: stdout differs\n";
-		print $fh "rx: $s\n";
-		return 0;
-	    }
+	    $self->transcript("stdout differs:\nrx: {$s}",
+			      locus => $self->{RUNCOM});
+	    $ok = 0;
 	}
     }
 
@@ -848,15 +1164,14 @@ sub runcom {
 	my $s = join("\n", @{$self->{RUNCOM}{stderr}});
 	if ($data{$child_stderr} !~ m{$s}ms) {
 	    $self->{failures}++;
-	    if (my $fh = $self->{xscript}) {
-		print $fh "$self->{filename}:$self->{RUNCOM}{BEG}-$self->{RUNCOM}{END}: stderr differs\n";
-		print $fh "rx: $s\n";
-		return 0;
-	    }
+	    $self->transcript("stderr differs:\nrx: {$s}",
+			      locus => $self->{RUNCOM});
+	    $ok = 0;
 	}
     }
 
-    return 1
+    $self->transcript($ok ? "OK" : "FAIL");
+    return $ok
 }
 
 sub parse_expect {
@@ -893,12 +1208,16 @@ sub parse_expect {
 sub replvar {
     my ($self, $var) = @_;
     if ($var =~ m{(LISTENER|BACKEND)(\d+)?(?::(PORT|IP))?}) {
-	my $var = ($1 eq 'LISTENER') ? $listeners : $backends;
+	my $list = ($1 eq 'LISTENER') ? $listeners : $backends;
 	my $meth = 'address';
 	if (defined($3)) {
 	    $meth = ($3 eq 'PORT') ? 'port' : 'host';
 	}
-	return $var->get($2//0)->${ \$meth };
+	if (my $obj = $list->get($2//0)) {
+	    return $obj->${ \$meth };
+	} else {
+	    $self->syntax_error("unrecognized variable $var");
+	}
     }
     return '${' . $var . '}';
 }
@@ -961,9 +1280,9 @@ sub parse_expect_body {
 	    return;
 	}
 
-	if (/^(?:#.*)?$/) {
-	    next;
-	}
+	# if (/^(?:#.*)?$/) {
+	#     next;
+	# }
 
 	if (/\\(.*)/) {
 	    push @{$self->{EXP}{BODY}}, $1;
@@ -1078,12 +1397,12 @@ if ($@) {
 }
 
 sub new {
-    my ($class, $number, $ident, $keepopen, $proto) = @_;
+    my ($class, $number, $ident, $keepopen, $proto, $ip) = @_;
     my $socket;
     socket($socket, PF_INET, SOCK_STREAM, 0)
 	or croak "socket: $!";
     setsockopt($socket, SOL_SOCKET, SO_REUSEADDR, 1);
-    bind($socket, pack_sockaddr_in(0, inet_aton('127.0.0.1')))
+    bind($socket, pack_sockaddr_in(0, inet_aton($ip // '127.0.0.1')))
 	or die "bind: $!";
     if ($keepopen) {
 	listen($socket, 128);
@@ -1093,12 +1412,12 @@ sub new {
 	    or croak "fcntl F_SETFD: $!";
     }
     my $sa = getsockname($socket);
-    my ($port, $ip) = sockaddr_in($sa);
+    my ($port, $ipaddr) = sockaddr_in($sa);
     bless {
 	number => $number,
 	ident => $ident,
 	socket => $socket,
-	host => inet_ntoa($ip),
+	host => inet_ntoa($ipaddr),
 	port => $port,
 	proto => lc($proto // "http")
     }, $class;
@@ -1132,7 +1451,8 @@ sub set_pass_fd {
 sub send_fd {
     my $self = shift;
     croak "not applicable" unless $self->pass_fd;
-    accept(my $fh, $self->pass_fd);
+    accept(my $fh, $self->pass_fd)
+	or croak "accept on pass_fd failed: $!";
     IO::FDPass::send(fileno($fh), $self->fd)
 	or croak "failed to pass socket: $!";
     close $self->socket_handle;
@@ -1160,7 +1480,7 @@ sub new {
 sub keepopen { shift->{keepopen} }
 sub count { scalar @{shift->{listeners}} }
 sub create {
-    my ($self, $ident, $proto) = @_;
+    my ($self, $ident, $proto, $ip) = @_;
     if (defined($proto) && lc($proto) eq 'HTTPS') {
 	my ($ok, $why) = HTTP::Tiny->can_ssl;
 	unless ($ok) {
@@ -1168,7 +1488,7 @@ sub create {
 	    exit(main::EX_SKIP);
 	}
     }
-    my $lst = Listener->new($self->count(), $ident, $self->keepopen, $proto);
+    my $lst = Listener->new($self->count(), $ident, $self->keepopen, $proto, $ip);
     if ($self->keepopen) {
 	$lst->set_pass_fd("lst".$self->count().".sock");
     }
@@ -1231,7 +1551,7 @@ sub read_and_process {
 	    $lst->listen;
 	    while (1) {
 		my $fh;
-		accept($fh, $lst->socket_handle);
+		accept($fh, $lst->socket_handle) or threads->exit();
 		process_http_request($fh, $lst)
 	    }
 	}, $lst)->detach;
@@ -1260,8 +1580,9 @@ sub http_redirect {
     my ($http, $rest) = @_;
     my $redir = $http->header('x-redirect')//'';
     $http->reply(301, "Moved Permanently", headers => {
-			'location' => $redir . '/echo' . $rest
-		 });
+	'location' => $redir . '/echo' . $rest,
+	'x-orig-location' => $redir . '/echo' . $rest
+    });
 }
 
 sub process_http_request {
@@ -1362,9 +1683,22 @@ sub ParseHeader {
 
 sub GetBody {
     my $http = shift;
-    if (my $len = $http->header('Content-Length')) {
+    if ($http->header('Transfer-Encoding')//'' eq 'chunked') {
+	my @chunks;
+	my $chunk_ext;
+	while (my $chunk_size_ext = $http->getline()) {
+	    (my $chunk_size, $chunk_ext) = split /;/, $chunk_size_ext, 2;
+	    my $len = hex($chunk_size);
+	    last if $len == 0;
+	    read($http->{fh}, my $chunk, $len);
+	    push @chunks, $chunk_size_ext, $chunk;
+	    read($http->{fh}, my $s, 2);
+	    # FIXME: error checking
+	}
+	$http->{BODY} = join("\n", @chunks);
+    } elsif (my $len = $http->header('Content-Length')) {
 	read($http->{fh}, $http->{BODY}, $len);
-    }
+    } # FIXME: else...
 }
 
 sub parse {
@@ -1390,6 +1724,95 @@ sub reply {
 	print $fh $opt{body};
     }
 }
+
+package PoundControl;
+use strict;
+use warnings;
+use Carp;
+use IPC::Open3;
+use IO::Select;
+use Symbol 'gensym';
+use Data::Dumper;
+
+sub new {
+    my ($class, $config) = @_;
+    return bless {
+	config => $config//'pound.cfg',
+    }, $class
+}
+
+sub request {
+    my ($self, $command, $arg) = @_;
+
+    local $SIG{CHLD} = sub {};
+    my ($child_stdin, $child_stdout, $child_stderr);
+    $child_stderr = gensym();
+    %status_codes = ();
+    my $pid = open3($child_stdin, $child_stdout, $child_stderr,
+		    'poundctl', '-f', $self->{config}, '-j', $command, $arg);
+    close $child_stdin;
+
+    my $sel = IO::Select->new();
+    $sel->add($child_stdout, $child_stderr);
+    my $CHUNK_SIZE = 1000;
+    my @ready;
+    my %data = ( $child_stdout => '', $child_stderr => '' );
+
+    while (@ready = $sel->can_read) {
+	foreach my $fh (@ready) {
+	    my $data;
+	    while (1) {
+		my $len = sysread($fh, $data{$fh}, $CHUNK_SIZE,
+				  length($data{$fh}));
+		croak "sysread: $!" unless defined($len);
+		if ($len == 0) {
+		    $sel->remove($fh);
+		    $fh->close;
+		    last;
+		}
+	    }
+	}
+    }
+
+    waitpid($pid, 0);
+
+    my $code;
+    if ($? == -1) {
+	croak "failed to run poundctl: $!";
+    } elsif ($? & 127) {
+	croak "poundctl terminated on signal ".($? & 127);
+    } elsif ($? >> 8) {
+	croak "poundctl failed: " . $data{$child_stderr};
+    }
+    return JSON->new->boolean_values(0,1)->decode($data{$child_stdout});
+}
+
+sub list {
+    my $self = shift;
+    my $arg;
+    if (@_) {
+	$arg .= '/' . join('/', @_)
+    }
+    return $self->request('list', $arg);
+}
+
+sub backends {
+    my ($self, $lst, $srv, %opts) = @_;
+    my $res = $self->list($lst, $srv);
+    if ($opts{sort}) {
+	$res->{backends} = [sort {
+	    if ($a->{weight} != $b->{weight}) {
+		$a->{weight} <=> $b->{weight}
+	    } elsif ($a->{priority} != $b->{priority}) {
+		$a->{priority} <=> $b->{priority}
+	    } else {
+		($a->{address}//'') cmp ($b->{address}//'')
+	    }
+	} @{$res->{backends}}];
+    }
+    return $res->{backends};
+}
+
 1;
 __END__
 
@@ -1401,12 +1824,17 @@ poundharness - run pound tests
 
 B<poundharness>
 [B<-sv>]
+[B<-a> I<IP>]
 [B<-f I<FILE>>]
 [B<-l I<N>>]
 [B<-t I<N>>]
 [B<-x I<FILE>>]
 [B<--config=>I<SRC>[B<:>I<DST>]]
+[B<--fakedns=>[I<PORT>:]I<LIB>]
+[B<--include-dir=>I<DIR>]
 [B<--log-level=>I<N>]
+[B<--preproc=>I<FILE>[:I<DST>]]
+[B<--source-address=>I<IP>]
 [B<--startup-timeout=>I<N>]
 [B<--statistics>]
 [B<--transcript=>I<FILE>]
@@ -1419,7 +1847,7 @@ Upon startup, B<poundharness> reads B<pound> configuration file F<pound.cfi>,
 modifies the settings as described below and writes the resulting configuration
 to B<pound.cfg>.  During modification, the following statements are removed:
 B<Daemon>, B<LogFacility> and B<LogLevel>.  The settings suitable for running
-B<pound> as a subordinate process are inserted istead.  In particular,
+B<pound> as a subordinate process are inserted instead.  In particular,
 B<LogLevel> is set from the value passed with the B<--log-level> command
 line option.  For each B<ListenHTTP> and B<ListenHTTPS> section, the actual
 socket configuration (i.e. the B<Address>, B<Port>, or B<SocketFrom> statements)
@@ -1430,6 +1858,9 @@ socket information is stored in the B<Address> and B<Port> statements that
 replace the removed ones and a backend HTTP server is started listening
 on that socket.
 
+A B<Control> statement is inserted at the beginning of the file, enabling
+the control interface via the F<pound.ctl> socket.
+
 B<Include> statements are processed in the following manner: the file
 supplied as the argument is preprocessed and the result is written to
 a file with the name obtained by appending suffix C<.pha> to the original
@@ -1476,11 +1907,52 @@ to I<DST> and use it as configuration file when running B<pound>.  If
 I<DST> is omitted, F<pound.cfg> is assumed.  If this option is not given,
 I<SRC> defaults to F<pound.cfi>.
 
+=item B<--fakedns=>I<LIB>
+
+Start a mock DNS server listening on arbitrary free ports (UDP and TCP),
+initialize environment variables B<FAKEDNS_UDP_PORT> and B<FAKEDNS_TCP_PORT>
+to these port numbers, and preload the library I<LIB> before exec'ing
+B<pound>.  I<LIB> must be the absolute pathname of the B<libfakedns.so> file.
+See B<fakedns.c> for details about this library.
+
+This option also instructs B<poundharness> to emit a B<Resolv> section to
+the created B<pound.cfg> file and to extend the script file syntax with
+the statements described in the B<DNS Statements> section (see below).
+
+If B<Net::DNS::Nameserver> perl module is not available and B<poundharness>
+is given this option, it will exit with status code 77.
+
+To avoid spurious failures, before actually using this functionality
+B<poundharness> performs a self-test, consisting in querying a SOA record of
+a mock DNS zone, using external helper program B<getsoa>.  If the test
+returns the correct data (indicating thereby that the trick with B<fakedns>
+works), B<poundharness> continues.  Otherwise, it exists with status code 77.
+
+=item B<--include-dir=>I<DIR>
+
+Look for relative file names in I<DIR>.  In particular, relative file names
+given as arguments to the B<--preproc> option or appearing in B<Include>
+statements in B<pound> configuration file will be searched in I<DIR>.
+This directory is also passed to B<pound> via the B<-Winclude-dir=I<DIR>>
+option.
+
 =item B<-l>, B<--log-level=> I<N>
 
 Set B<pound> I<LogLevel> configuration parameter.  If I<N> is B<-1>,
 I<LogLevel> set in the configuration file is used.
 
+=item B<--preproc=>I<FILE>[:I<DST>]
+
+Preprocess I<FILE> as described in the B<DESCRIPTION>.  Write the resulting
+material to I<DST>.  If the latter is omitted, the destination file will be
+named I<FILE>B<.cfg>.
+
+=item B<-a>, B<--source-address=> I<IP>
+
+Register I<IP> as a possible source address for HTTP requests.  Argument
+should match C<127.0.0.0/8>.  Registered source addresses can be used in
+B<source> statement in the script file.
+
 =item B<-s>, B<--statistics>
 
 Print short statistics at the end of the run.
@@ -1516,6 +1988,10 @@ A request starts with a line specifying the request method in uppercase
 by arbitrary number of HTTP headers, newline and request body.  The request
 is terminated with the word B<end> on a line alone.
 
+The sequence B<@@> has a special meaning when used in the body.  These
+characters instruct B<poundharness> to send the body using B<chunked> transfer
+coding and indicate the beginning of each chunk.
+
 An expected response must follow each request.  It begins with a
 three digit response code on a line alone.  The code may be followed by
 any number of response headers.  If present, the test will succeed only
@@ -1580,6 +2056,17 @@ correspondingly.
 
 Numbering of listeners and backends starts from 0.
 
+By default, requests sent by B<poundharness.pl> originate from 127.0.0.1.
+Another source address can be set using the B<source> statement:
+
+  source IP
+
+The IP must have been previously declared as a source using the
+B<--source-address> command line option.  Obviously, the networking
+configuration of the host machine must allow it to be used as a
+source IP, otherwise B<poundharness.pl> issue a diagnostic message and
+exit with code 77.
+
 The B<server> statement can be used to specify the B<pound> listener to
 send the requests to.  It has the form
 
@@ -1593,7 +2080,7 @@ occurs first.
 =head2 External program send/expect
 
 This type of stanza allow you to run an external program and examine its
-exit code, standard output and error streams.  It is inlcuded mainly for
+exit code, standard output and error streams.  It is included mainly for
 testing the B<poundctl> command.
 
 The stanza begins with the keyword B<run> followed by the command
@@ -1609,7 +2096,7 @@ Expect program to return exit status I<N> (a decimal number).
 
 Expect text on stdout.  Everything below this keyword and up to the
 B<end> keyword appearing on a line alone is taken to be the expected
-text.  When matching actiual program output, this text is treated as
+text.  When matching actual program output, this text is treated as
 Perl multi-line regular expression (see the B<m> and B<s> flags in
 B<perlre>).  To expect a line containing the word C<end> alone, prefix
 it with a backslash.
@@ -1621,6 +2108,118 @@ its syntax.
 
 =back
 
+=head2 Backend Usage Analysis
+
+The statement
+
+=over 4
+
+=item B<stats> B<samples>=I<N> I<K>=I<V>...
+
+=back
+
+causes the following send/expect statements to be executed I<N> times in
+turn.  After this, backend usage statistics will be computed and the
+resulting values compared with the expected values supplied by I<K>=I<V>
+pairs.  Allowed values for I<K> are:
+
+=over 4
+
+=item B<samples>
+
+Number of samples to run.
+
+=item B<min>
+
+Minimum number of requests served by a backend.
+
+=item B<max>
+
+Maximum number of requests served by a backend.
+
+=item B<avg>
+
+Average number of requests served by a backend.
+
+=item B<stddev>
+
+Standard deviation of the number of requests served by a backend.
+
+=item B<index>
+
+Apply the above values to this backend (0-based index).
+
+=back
+
+=head2 Querying the Control Interface
+
+The following statement query the B<pound> control interface and verify
+if the response matches the expectation.  The latter is supplied after the
+statement, in JSON format, and ends with the keyword B<end> on a line by
+itself.  Only attributes present in the expectation are compared.
+
+=over 4
+
+=item backends I<LN> I<SN>
+
+Query for backends in service I<SN> of listener I<LN>.
+The returned backend array is sorted by weight, priority and address.
+
+=back
+
+=head2 Additional Directives
+
+=over 4
+
+=item B<mkbackend> I<IP>
+
+Declares new backend with the given IP address.  I<IP> must fall within
+127.0.0.0/8.
+
+Normally backends are created automatically and this statement is not needed.
+Use it for testing dynamic backends.
+
+=item B<sleep> I<N>
+
+Pause execution for I<N> seconds.
+
+=item B<echo> I<WORD> [I<WORD>...]
+
+Expand variables (see B<HTTP send/expect> above) in I<WORD>s and output
+the result on standard error, prefixed with the B<#> sign.
+
+=head2 DNS Statements
+
+The following keywords are designed for testing dynamic DNS-based backends.
+They become available if the following two conditions are met:
+
+=over 4
+
+=item *
+
+The module B<Net::DNS::Nameserver> is available.
+
+=item *
+
+Full pathname of the B<libfakedns.so> library is given using the
+B<--fakedns> option.
+
+=back
+
+When these conditions are met, B<poundharness> will start a subsidiary DNS
+server and modify the produced B<pound.cfg> file to use it.  The following
+extended keywords can then be used:
+
+=over 4
+
+=item zonefile
+
+This statement defines the DNS zone (or zones) served by the subsidiary
+DNS.  The zone definition follows the keyword and ends with the keyword
+B<end> on a line by itself.  Zone file syntax is defined by RFC1035.
+
+=back
+
 =head1 BACKENDS
 
 Each B<Backend> statement in the configuration file causes creation of
@@ -1653,6 +2252,9 @@ The value of the header I<header> in the request.
 
 =back
 
+If the body was sent using B<chunked> encoding, it is reproduced
+verbatim, instead of reconstructing it as per RFC 9112, 7.1.3.
+
 =head2 /redirect
 
 Redirects the request to the B</echo> endpoint.  The value of the
